""" QuickMPCからのレスポンスに含まれる計算処理の進捗を表示するデモ
本デモコードは次の操作を行っている．
Step 1. QuickMPC動作のための準備
Step 2. 送信用データの生成
Step 3. シェアをコンテナに送信
Step 4. 計算リクエストを送信
Step 5. レスポンスからプログレスバーを作成・更新
"""

import logging
import time
import collections
from typing import List, Tuple, OrderedDict, Optional

import tqdm

from quickmpc import QMPC, JobStatus

logger = logging.getLogger(__name__)
formatter = ("%(levelname)-5s | %(message)s")
logging.basicConfig(level=logging.DEBUG, format=formatter)


def large_data(size: int, data_num: int):
    schema_size: int = 10
    schema: List[str] = ["id"] + \
        [f"s{i}_{data_num}" for i in range(schema_size - 1)]

    mod = 47
    data: List[List[str]] = [
        [str(i)] + [str(((i * size + k) % mod) / mod)
                    for k in range(schema_size - 1)]
        for i in range(size)
    ]
    return qmpc.parse_csv_data([schema] + data)


def get_data_id(size: int, data_num):
    """ Step 2. 送信用データの生成 """
    secrets, schema = large_data(size, data_num)

    """ Step 3. シェアをコンテナに送信 """
    res = qmpc.send_share(secrets, schema)
    assert (res["is_ok"])
    data_id: str = res["data_id"]

    return (data_id, secrets, schema)


if __name__ == '__main__':

    """ Step 1. 諸準備 """
    # QuickMPC Setting
    qmpc: QMPC = QMPC(
        ["http://localhost:9001",
         "http://localhost:9002",
         "http://localhost:9003"]
    )
    # データ Setting
    size = 50000
    data_id1, secrets1, schema1 = get_data_id(size, data_num=1)
    data_id2, secrets2, schema2 = get_data_id(size, data_num=2)

    """ Step 4. テーブル横結合をリクエスト"""
    table = ([data_id1, data_id2], [2], [1, 1])
    exec_res = qmpc.get_join_table(table)

    job_uuid = exec_res["job_uuid"]
    logger.info(f"job_uuid: {job_uuid}")

    pbars: OrderedDict[Tuple[int, int],
                       Tuple[Optional[tqdm.tqdm], float]] \
        = collections.OrderedDict()

    QMPC.set_log_level(logging.WARN)

    while True:
        get_res = qmpc.get_computation_result(job_uuid)

        """ Step 5. プログレスバーの作成・更新 """
        if get_res['statuses'] is not None:
            for party_id, status in enumerate(get_res['statuses']):
                key = (party_id, -1)

                if key not in pbars:
                    pbars[(party_id, -1)] = (tqdm.tqdm(
                        desc=f"[{party_id}] status",
                        total=len(JobStatus.items()) - 1), 0)

                pbar, prev = pbars[key]
                if pbar is None:
                    continue

                pbar.update(status - prev)
                pbar.set_postfix(
                    status=JobStatus.Name(status)
                )

                if status == JobStatus.Value('COMPLETED'):
                    pbar.close()
                    pbar = None

                pbars[key] = (pbar, status)

        progresses = get_res['progresses']
        if progresses is not None:
            for party_id, progress in enumerate(progresses):
                if progress is None:
                    continue

                for procedure in progress.progresses:
                    key = (party_id, procedure.id)

                    if key not in pbars:
                        pbars[key] = (tqdm.tqdm(
                            desc=f"[{party_id}] {procedure.description}",
                            total=100), 0)

                    pbar, prev = pbars[key]
                    if pbar is None:
                        continue

                    pbar.update(procedure.progress - prev)
                    pbar.set_postfix(details=procedure.details)

                    if procedure.completed:
                        pbar.close()
                        pbar = None

                    pbars[key] = (pbar, procedure.progress)

        if get_res["results"] is not None:
            for key in pbars:
                pbar, prev = pbars[key]
                if pbar is None:
                    continue
                pbar.update(100 - prev)
                pbar.clear()
                pbar.close()

            break

        time.sleep(1)



import functools
import operator
from dataclasses import dataclass
from typing import List

from ..exception import ArgmentError
from .overload_tools import Dim1, Dim2, methoddispatch


@dataclass(frozen=True)
class MakePiece:

    @methoddispatch(is_static_method=True)
    @staticmethod
    def __get_byte(p): ...

    @__get_byte.register(float)
    @__get_byte.register(int)
    @staticmethod
    def __get_byte_number(f):
        # 扱う数値は64bitなので8byte
        return 8

    @__get_byte.register(str)
    @staticmethod
    def __get_byte_str(s: str):
        return len(s)

    @methoddispatch(is_static_method=True)
    @staticmethod
    def make_pieces(_, __):
        raise ArgmentError("不正な引数が与えられています．")

    @staticmethod
    def check_max_size(max_size: int):
        # NOTE: Couchbaseのアイテムサイズ上限:1MB
        # NOTE: grpcの送受信データサイズ上限:4MB
        lower_limit_size: int = 1
        upper_limit_size: int = 1_000_000
        if max_size > upper_limit_size:
            raise RuntimeError(
                f"max size of piece ({max_size}) is over "
                f"specified limit size ({upper_limit_size})"
            )
        if max_size < lower_limit_size:
            raise RuntimeError(
                f"max size of piece ({max_size}) is under "
                f"specified limit size ({lower_limit_size})"
            )

    @make_pieces.register(Dim1)
    @staticmethod
    def __make_pieces_1d(src: List[str], max_size: int) -> List[List[str]]:
        MakePiece.check_max_size(max_size)
        cur_size = 0
        index = 0
        dst: List[List[str]] = [[]]
        for val in src:
            val_size = MakePiece.__get_byte(val)
            if val_size > max_size:
                raise RuntimeError(
                    f"line size ({val_size}) is over "
                    f"specified limit size ({max_size})"
                )
            cur_size += val_size
            if cur_size > max_size:
                cur_size = val_size
                index += 1
                dst.append([])

            dst[index].append(val)

        return dst

    @make_pieces.register(Dim2)
    @staticmethod
    def __make_pieces_2d(src: List[List[str]],
                         max_size: int
                         ) -> List[List[List[str]]]:
        MakePiece.check_max_size(max_size)
        cur_size = 0
        index = 0
        dst: List[List[List[str]]] = [[]]
        for row in src:
            row_size = functools.reduce(
                operator.add, map(MakePiece.__get_byte, row), 0)
            if row_size > max_size:
                raise RuntimeError(
                    f"line size ({row_size}) is over "
                    f"specified limit size ({max_size})"
                )
            cur_size += row_size

            if cur_size > max_size:
                cur_size = row_size
                index += 1
                dst.append([])

            dst[index].append(row)

        return dst

    @make_pieces.register(str)
    @staticmethod
    def __make_pieces_str(src: str, max_size: int) -> List[str]:
        MakePiece.check_max_size(max_size)
        dst: List[str] = []
        current: str = ""
        for ch in src:
            current += ch

            if len(current) == max_size:
                dst.append(current)
                current = ""
        if current:
            dst.append(current)
        return dst



from abc import ABC, abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from typing import ClassVar, List

from nacl.utils import random, randombytes_deterministic

from .overload_tools import methoddispatch
from ..exception import ArgmentError


# 乱数生成のインタフェース
class RandomInterface(ABC):
    @abstractmethod
    def get(self, a, b) -> int:
        ...

    @abstractmethod
    def get_list(self, a, b, size: int) -> List[int]:
        ...


@dataclass(frozen=True)
class ChaCha20(RandomInterface):

    # 128bit符号付き整数最大，最小値
    mx: ClassVar[int] = (1 << 128)-1
    mn: ClassVar[int] = -(1 << 128)

    @methoddispatch()
    def get(self, a, b):
        raise ArgmentError(
            "乱数の閾値はどちらもintもしくはdecimalでなければなりません．"
            f"a is {type(a)}, b is {type(b)}")

    @get.register(int)
    def __get_int(self, a: int, b: int) -> int:
        # TRNGで32byte(256bit)生成
        self.__exception_check(a, b)
        byte_val: bytes = self.__get_32byte()
        int32byte = int.from_bytes(byte_val, "big")
        return int32byte % (b-a)+a

    @get.register(Decimal)
    def __get_decimal(self, a: Decimal, b: Decimal) -> Decimal:
        # 256bit整数を取り出して[a,b]に正規化する
        self.__exception_check(a, b)
        val: int = self.get(self.mn, self.mx)
        return Decimal(val-self.mn)/(self.mx-self.mn)*(b-a)+a

    @methoddispatch()
    def get_list(self, a, b, size: int):
        raise ArgmentError(
            "乱数の閾値はどちらもintもしくはdecimalでなければなりません．"
            f"a is {type(a)}, b is {type(b)}")

    @get_list.register(int)
    def __get_list_int(self, a: int, b: int, size: int) -> List[int]:
        # TRNGの32byteをseedとしてCSPRNGでsize分生成
        byte_size: int = 16
        self.__exception_check(a, b)
        seed: bytes = self.__get_32byte()
        bytes_list: bytes = randombytes_deterministic(size*byte_size, seed)
        int_list = [int.from_bytes(bytes_list[i:i+byte_size], "big")
                    for i in range(0, len(bytes_list), byte_size)]
        return [x % (b-a)+a for x in int_list]

    @get_list.register(Decimal)
    def __get_list_decimal(self, a: Decimal, b: Decimal, size: int) \
            -> List[Decimal]:
        # 128bit整数を取り出して[a,b]に正規化する
        self.__exception_check(a, b)
        valList: List[int] = self.get_list(self.mn, self.mx, size)
        return [Decimal(val-self.mn)/(self.mx-self.mn)*(b-a)+a
                for val in valList]

    def __get_32byte(self) -> bytes:
        return random()

    def __exception_check(self, a, b) -> None:
        if a >= b:
            raise ArgmentError(
                "乱数の下限は上限より小さい必要があります．"
                f"{a} < {b}")
        if type(a) != type(b):
            raise ArgmentError(
                "乱数の下限と上限の型は一致させる必要があります．"
                f"{type(a)} != {type(b)}")



from typing import Any, Callable, Optional


def if_present(optional: Optional[Any],
               func: Callable[[Any], Any],
               *arg: Any
               ) -> Optional[Any]:
    if optional is None:
        return None
    return func(optional, *arg)



try:
    from ._version import __version__  # type: ignore
except ImportError:
    try:
        from setuptools_scm import get_version  # type: ignore
        __version__ = get_version(root='../../../', relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"




import logging

from quickmpc import QMPC

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

if __name__ == '__main__':
    # QuickMPC Setting
    qmpc: QMPC = QMPC(
        ["http://localhost:9001",
         "http://localhost:9002",
         "http://localhost:9003"]
    )

    # job_uuidを指定して計算結果を取得する
    job_uuid: str = "10b6bfda-8776-4d8c-ad9a-b40fa277ff37"
    res = qmpc.get_computation_result(job_uuid)
    secrets = res["results"]
    logger.info(secrets)





import logging

from quickmpc import QMPC

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

if __name__ == '__main__':
    # QuickMPC Setting
    qmpc: QMPC = QMPC(
        ["http://localhost:9001",
         "http://localhost:9002",
         "http://localhost:9003"]
    )

    # 複数のシェアを作成・送信
    data_ids = []
    for i in range(3):
        res = qmpc.send_share([[1]], ["key"])
        data_ids.append(res['data_id'])
        logger.info(res)

    # シェアを削除
    for i in range(2):
        res = qmpc.delete_share(data_ids)
        logger.info(res)
        # 2 回目は失敗することを確認
        assert res["is_ok"] == (i == 0)



import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import ClassVar, List, Tuple, Callable, Any

import numpy as np

from .utils.overload_tools import (DictList, DictList2,
                                   Dim1, Dim2, Dim3, methoddispatch)
from .utils.random import ChaCha20, RandomInterface
from .exception import ArgmentError

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Share:
    __share_random_range: ClassVar[Tuple[Decimal, Decimal]] =\
        (Decimal(-(1 << 64)), Decimal(1 << 64))

    @staticmethod
    def __to_str(val: Decimal) -> str:
        # InfinityをCCで読み込めるinfに変換
        return 'inf' if Decimal.is_infinite(val) else str(val)

    @methoddispatch(is_static_method=True)
    @staticmethod
    def sharize(_, __):
        logger.error("Invalid argument on sharize.")
        raise ArgmentError("不正な引数が与えられています．")

    @methoddispatch(is_static_method=True)
    @staticmethod
    def recons(_):
        logger.error("Invalid argument on recons.")
        raise ArgmentError("不正な引数が与えられています．")

    @sharize.register(int)
    @sharize.register(float)
    @staticmethod
    def __sharize_scalar(secrets: float, party_size: int = 3) -> List[str]:
        """ スカラ値のシェア化 """
        rnd: RandomInterface = ChaCha20()
        shares: List[int] = rnd.get_list(
            *Share.__share_random_range, party_size)
        shares[0] += Decimal(secrets) - np.sum(shares)
        shares_str: List[str] = [str(n) for n in shares]
        return shares_str

    @sharize.register(Dim1)
    @staticmethod
    def __sharize_1dimension(secrets: List[float], party_size: int = 3) \
            -> List[List[str]]:
        """ 1次元リストのシェア化 """
        rnd: RandomInterface = ChaCha20()
        secrets_size: int = len(secrets)
        shares: np.ndarray = np.array([
            rnd.get_list(*Share.__share_random_range, secrets_size)
            for __ in range(party_size - 1)])
        s1: np.ndarray = np.subtract(np.frompyfunc(Decimal, 1, 1)(secrets),
                                     np.sum(shares, axis=0))
        shares_str: List[List[str]] = \
            np.vectorize(Share.__to_str)([s1, *shares]).tolist()
        return shares_str

    @sharize.register(Dim2)
    @staticmethod
    def __sharize_2dimension(secrets: List[List[float]],
                             party_size: int = 3) -> List[List[List[str]]]:
        """ 2次元リストのシェア化 """
        rnd: RandomInterface = ChaCha20()
        secrets_size: int = len(secrets)
        secrets_size2: int = len(secrets[0])
        shares: np.ndarray = np.array([
            [rnd.get_list(*Share.__share_random_range, secrets_size2)
             for __ in range(secrets_size)]
            for ___ in range(party_size - 1)
        ])
        s1: np.ndarray = np.subtract(
            np.frompyfunc(Decimal, 1, 1)(np.array(secrets, dtype=Decimal)),
            np.sum(shares, axis=0))
        shares_str: List[List[List[str]]] = \
            np.vectorize(Share.__to_str)([s1, *shares]).tolist()
        return shares_str

    @sharize.register(dict)
    @staticmethod
    def __sharize_dict(secrets: dict, party_size: int = 3) -> List[dict]:
        """ 辞書型のシェア化 """
        shares_str: List[dict] = [dict() for _ in range(party_size)]
        for key, val in secrets.items():
            for i, share_val in enumerate(Share.sharize(val, party_size)):
                shares_str[i][key] = share_val
        return shares_str

    @sharize.register(DictList)
    @staticmethod
    def __sharize_dictlist(secrets: dict, party_size: int = 3) \
            -> List[List[dict]]:
        """ 辞書型配列のシェア化 """
        shares_str: List[List[dict]] = [[] for _ in range(party_size)]
        for secret_dict in secrets:
            share_dict: List[dict] = Share.sharize(secret_dict, party_size)
            for ss, sd in zip(shares_str, share_dict):
                ss.append(sd)
        return shares_str

    @recons.register(Dim1)
    @staticmethod
    def __recons_list1(shares: List, f: Callable[[Any], Any] = float):
        """ 1次元リストのシェアを復元 """
        try:
            # 文字列がfloatかどうかcheck
            [float(x) for x in shares]
        except ValueError:
            return shares[0]
        return f(sum([Decimal(x) for x in shares]))

    @recons.register(Dim2)
    @recons.register(Dim3)
    @staticmethod
    def __recons_list(shares: List[List],
                      f: Callable[[Any], Any] = float) -> List:
        """ リストのシェアを復元 """
        secrets: List = [
            Share.recons([shares_pi[i] for shares_pi in shares], f)
            for i in range(len(shares[0]))
        ]
        return secrets

    @recons.register(DictList)
    @staticmethod
    def __recons_dictlist(shares: List[dict],
                          f: Callable[[Any], Any] = float) -> dict:
        """ 辞書型を復元 """
        secrets: dict = dict()
        for key in shares[0].keys():
            val = []
            for s in shares:
                val.append(s[key])
            secrets[key] = Share.recons(val, f)
        return secrets

    @recons.register(DictList2)
    @staticmethod
    def __recons_dictlist2(shares: List[List[dict]],
                           f: Callable[[Any], Any] = float) -> list:
        """ 辞書型配列を復元 """
        secrets: list = list()
        for i in range(len(shares[0])):
            val = []
            for s in shares:
                val.append(s[i])
            secrets.append(Share.recons(val, f))
        return secrets



import csv
import datetime
import hashlib
import json
import logging
import os
import struct
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field, InitVar
from typing import Any, Dict, Iterable, List, Optional, Tuple
from urllib.parse import urlparse

import grpc
import numpy as np
import tqdm  # type: ignore
from grpc_status import rpc_status  # type: ignore

from .exception import ArgmentError, QMPCJobError, QMPCServerError
from .proto.common_types.common_types_pb2 import JobErrorInfo, JobStatus
from .proto.libc_to_manage_pb2 import (DeleteSharesRequest,
                                       ExecuteComputationRequest,
                                       GetComputationResultRequest,
                                       GetComputationResultResponse,
                                       GetDataListRequest,
                                       GetElapsedTimeRequest,
                                       GetJobErrorInfoRequest, Input,
                                       JoinOrder, SendSharesRequest)
from .proto.libc_to_manage_pb2_grpc import LibcToManageStub
from .share import Share
from .utils.if_present import if_present
from .utils.make_pieces import MakePiece
from .utils.overload_tools import Dim2, Dim3, methoddispatch
from .utils.parse_csv import format_check

abs_file = os.path.abspath(__file__)
base_dir = os.path.dirname(abs_file)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class QMPCServer:
    endpoints: InitVar[List[str]]
    __client_stubs: Tuple[LibcToManageStub] = field(init=False)
    __party_size: int = field(init=False)
    token: str

    def __post_init__(self, endpoints: List[str]) -> None:
        stubs = [LibcToManageStub(QMPCServer.__create_grpc_channel(ep))
                 for ep in endpoints]
        object.__setattr__(self, "_QMPCServer__client_stubs", stubs)
        object.__setattr__(self, "_QMPCServer__party_size", len(endpoints))

    @staticmethod
    def __create_grpc_channel(endpoint: str) -> grpc.Channel:
        channel: grpc.Channel = None
        o = urlparse(endpoint)
        if o.scheme == 'http':
            # insecureなchannelを作成
            channel = grpc.insecure_channel(o.netloc)
        elif o.scheme == 'https':
            # secureなchannelを作成
            credential: grpc.ChannelCredentials \
                = grpc.ssl_channel_credentials()
            channel = grpc.secure_channel(o.netloc, credential)
        else:
            logger.error(f'仕様を満たさない形式のendpointが渡された: {endpoint}')
            raise ArgmentError(
                "endpointsにサポートされてないプロトコルが指定されています．http/httpsのいずれかを指定してください．")

        return channel

    @staticmethod
    def _argument_check(join_order: Tuple[List, List, List]):
        if len(join_order[0])-1 != len(join_order[1]):
            logger.error(
                'the size of join must be one less than the size of dataIds')
            return False
        if len(join_order[0]) != len(join_order[2]):
            logger.error('the size of index must match the size of dataIds')
            return False
        # TODO joinをenumにする
        if not all([0 <= join <= 2 for join in join_order[1]]):
            logger.error('join value must be in the range of 0 to 2')
            return False
        return True

    @staticmethod
    def __futures_result(
            futures: Iterable, enable_progress_bar=True) -> Tuple[bool, List]:
        """ エラーチェックしてfutureのresultを得る """
        is_ok: bool = True
        response: List = []
        try:
            if enable_progress_bar:
                futures = tqdm.tqdm(futures, desc='receive')
            response = [f.result() for f in futures]
        except grpc.RpcError as e:
            is_ok = False
            logger.error(f'{e.details()} ({e.code()})')

            # エラーが詳細な情報を持っているか確認
            status = rpc_status.from_call(e)
            if status is not None:
                for detail in status.details:
                    if detail.Is(
                        JobErrorInfo.DESCRIPTOR   # type: ignore[attr-defined]
                    ):
                        # CC で Job 実行時にエラーが発生していた場合
                        # 例外を rethrow する
                        err_info = JobErrorInfo()
                        detail.Unpack(err_info)
                        logger.error(f"job error information: {err_info}")

                        raise QMPCJobError(err_info) from e

            # MC で Internal Server Error が発生している場合
            # 例外を rethrow する
            if e.code() == grpc.StatusCode.UNKNOWN:
                raise QMPCServerError("backend server return error") from e
        except Exception as e:
            is_ok = False
            logger.error(e)

        for b in response:
            if hasattr(b, "is_ok"):
                is_ok &= b.is_ok
            else:
                is_ok &= b["is_ok"]
        return is_ok, response

    @staticmethod
    def __stream_result(stream: Iterable, job_uuid: str, party: int,
                        path: Optional[str]) -> Dict:
        """ エラーチェックしてstreamのresultを得る """
        is_ok: bool = True
        res_list = []
        for res in stream:
            is_ok &= res.is_ok
            if path is not None:
                file_title = "dim1"
                if res.HasField("is_dim2"):
                    file_title = "dim2"
                elif res.HasField("is_schema"):
                    file_title = "schema"

                file_path = f"{path}/" + \
                    f"{file_title}-{job_uuid}-{party}-{res.piece_id}.csv"

                with open(file_path, 'w') as f:
                    writer = csv.writer(f)
                    writer.writerow([res.column_number])
                    writer.writerow(res.result)
                progress = res.progress if res.HasField('progress') else None
                res = GetComputationResultResponse(
                    message=res.message,
                    is_ok=res.is_ok,
                    column_number=res.column_number,
                    status=res.status,
                    piece_id=res.piece_id,
                    progress=progress,
                )
            res_list.append(res)
        res_dict: Dict = {"is_ok": is_ok, "responses": res_list}
        return res_dict

    @methoddispatch()
    def send_share(self, _):
        raise ArgmentError("不正な引数が与えられています．")

    @send_share.register(Dim2)
    @send_share.register(Dim3)
    def __send_share_impl(self, secrets: List, schema: List[str],
                          matching_column: int,
                          piece_size: int) -> Dict:
        if piece_size < 1000 or piece_size > 1_000_000:
            raise RuntimeError(
                "piece_size must be in the range of 1000 to 1000000")

        if matching_column <= 0 or matching_column > len(schema):
            raise RuntimeError(
                "matching_column must be in the "
                "range of 1 to the size of schema")

        # TODO parse_csv経由でsend_shareをすると同じチェックをすることになる．
        if not format_check(secrets, schema):
            raise RuntimeError("規定されたフォーマットでないデータです．")

        """ Shareをコンテナに送信 """
        sorted_secrets = sorted(
            secrets, key=lambda row: row[matching_column - 1])
        # pieceに分けてシェア化
        pieces: list = MakePiece.make_pieces(
            sorted_secrets, int(piece_size / 10))
        data_id: str = hashlib.sha256(
            str(sorted_secrets).encode() + struct.pack('d', time.time())
        ).hexdigest()
        shares = [Share.sharize(s, self.__party_size)
                  for s in tqdm.tqdm(pieces, desc='sharize')]
        sent_at = str(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

        # リクエストパラメータを設定して非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(stub.SendShares,
                                   SendSharesRequest(
                                       data_id=data_id,
                                       shares=json.dumps(s),
                                       schema=schema,
                                       piece_id=piece_id,
                                       sent_at=sent_at,
                                       matching_column=matching_column,
                                       token=self.token))
                   for piece_id, share_piece in enumerate(shares)
                   for stub, s in zip(self.__client_stubs, share_piece)
                   ]
        is_ok, _ = QMPCServer.__futures_result(futures)
        return {"is_ok": is_ok, "data_id": data_id}

    def delete_share(self, data_ids: List[str]) -> Dict:
        """ Shareを削除 """
        req = DeleteSharesRequest(dataIds=data_ids, token=self.token)
        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(stub.DeleteShares, req)
                   for stub in self.__client_stubs]
        is_ok, _ = QMPCServer.__futures_result(futures)
        return {"is_ok": is_ok}

    def execute_computation(self, method_id: int,
                            join_order: Tuple[List, List, List],
                            inp: Tuple[List, List]) -> Dict:
        if not self._argument_check(join_order):
            raise ArgmentError("引数が正しくありません．")

        """ 計算リクエストを送信 """
        join_order_req = JoinOrder(
            dataIds=join_order[0],
            join=join_order[1],
            index=join_order[2])
        input_req = Input(
            src=inp[0],
            target=inp[1])
        req = ExecuteComputationRequest(
            method_id=method_id,
            token=self.token,
            table=join_order_req,
            arg=input_req,
        )

        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        # JobidをMCから貰う関係で単一MC（現在はSP（ID=0）のみ対応）にリクエストを送る
        futures = [executor.submit(
            self.__client_stubs[0].ExecuteComputation, req)]

        is_ok, response = QMPCServer.__futures_result(futures)
        job_uuid = response[0].job_uuid if is_ok else None

        return {"is_ok": is_ok, "job_uuid": job_uuid}

    def get_data_list(self) -> Dict:
        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(stub.GetDataList,
                                   GetDataListRequest(token=self.token))
                   for stub in self.__client_stubs]
        is_ok, response = QMPCServer.__futures_result(futures)
        results = [eval(r.result) for r in response] if is_ok else None

        return {"is_ok": is_ok, "results": results}

    def get_elapsed_time(self, job_uuid: str) -> Dict:
        # リクエストパラメータを設定
        req = GetElapsedTimeRequest(
            job_uuid=job_uuid,
            token=self.token
        )
        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(stub.GetElapsedTime,
                                   req)
                   for stub in self.__client_stubs]
        is_ok, response = QMPCServer.__futures_result(futures)
        elapsed_time = max([res.elapsed_time
                            for res in response]) if is_ok else None
        return {"is_ok": is_ok, "elapsed_time": elapsed_time}

    def get_computation_result(self, job_uuid: str,
                               path: Optional[str]) -> Dict:
        """ コンテナから結果を取得 """
        # リクエストパラメータを設定
        req = GetComputationResultRequest(
            job_uuid=job_uuid,
            token=self.token
        )
        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(QMPCServer.__stream_result,
                                   stub.GetComputationResult(req),
                                   job_uuid, party, path)
                   for party, stub in enumerate(self.__client_stubs)]
        is_ok, response = QMPCServer.__futures_result(
            futures, enable_progress_bar=False)
        results_sorted = [sorted(res["responses"], key=lambda r: r.piece_id)
                          for res in response]
        # NOTE: statusは0番目(piece_id=1)の要素にのみ含まれている
        statuses = [res[0].status for res in results_sorted] \
            if results_sorted else None
        all_completed = all([
            s == JobStatus.Value('COMPLETED') for s in statuses
        ]) if statuses is not None else False

        progresses = None
        if results_sorted is not None:
            progresses = [
                res[0].progress if res[0].HasField("progress") else None
                for res in results_sorted
            ]

        results: Optional[Any] = None
        if not path and all_completed:
            for res in results_sorted:
                is_table = False
                is_dim2 = False
                column_number = 0
                result: Any = []
                schema = []
                for r in res:
                    if r.HasField("is_schema"):
                        if not is_table:
                            is_table = True
                        for val in r.result:
                            schema.append(val)
                    else:
                        if r.HasField("is_dim2"):
                            is_dim2 = True
                        for val in r.result:
                            result.append(val)

                    column_number = r.column_number

                if is_dim2:
                    result = np.array(result).reshape(-1,
                                                      column_number).tolist()
                result = {"schema": schema, "table": result} if is_table \
                    else result
                if results is None:
                    results = []
                results.append(result)

        results = if_present(results, Share.recons)
        return {"is_ok": is_ok, "statuses": statuses,
                "results": results, "progresses": progresses}

    def get_job_error_info(self, job_uuid: str) -> Dict:
        # リクエストパラメータを設定
        req = GetJobErrorInfoRequest(
            job_uuid=job_uuid,
            token=self.token
        )
        # 非同期にリクエスト送信
        executor = ThreadPoolExecutor()
        futures = [executor.submit(stub.GetJobErrorInfo, req)
                   for stub in self.__client_stubs]
        is_ok, response = QMPCServer.__futures_result(
            futures, enable_progress_bar=False)

        job_error_info = [
            res.job_error_info if res.HasField("job_error_info") else None
            for res in response
        ]

        return {"is_ok": is_ok, "job_error_info": job_error_info}



from decimal import Decimal
from typing import List

import numpy as np
import pytest

from quickmpc.share import Share


class TestQMPC:

    @pytest.mark.parametrize(
        ("secrets"),
        [
            # 1次元配列の秘密情報
            ([1]), ([3.0, 5.12, 21.05]), ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            # 2次元配列の秘密情報
            ([[1]]), ([[3000, 500, 2100], [100, 2500, 1100]]), ([[2.1, 3.2]]),
            ([[1, 2, 3, 4, 5],
              [6, 7, 8, 9, 10],
              [11, 12, 13, 14, 15],
              [16, 17, 18, 19, 20],
              [21, 22, 23, 24, 25]])
        ]
    )
    def test_sharize(self, secrets: list):
        """ nパーティのシェア化が正しくできているかのTest """
        for party_size in range(2, 10):
            shares: np.ndarray = np.vectorize(Decimal)(
                Share.sharize(secrets, party_size=party_size))
            assert (len(shares) == party_size)
            assert (np.allclose(secrets,
                                np.vectorize(float)(np.sum(shares, axis=0))))

    @pytest.mark.parametrize(
        ("secrets"),
        [(1), (0.4), (-100), (-50.14)]
    )
    def test_sharize_scalar(self, secrets: list):
        """ nパーティのスカラ値のシェア化が正しくできているかのTest """
        for party_size in range(2, 10):
            shares: np.ndarray = np.vectorize(Decimal)(
                Share.sharize(secrets, party_size=party_size))
            assert (len(shares) == party_size)
            assert (float(np.sum(shares)) == pytest.approx(secrets))

    @pytest.mark.parametrize(
        ("secrets"),
        [
            ([{"a": 3, "b": 2}]),
            ([{"a": 3, "b": {"b1": 1, "b2": 10}}]),
            ([{"a": 3, "b": {"b1": 1, "b2": {"b21": {"b211": 100}}}},
              {"c": 3, "d": {"d1": 1, "d2": {"d21": {"d211": 100}}}}]),
        ]
    )
    def test_sharize_json(self, secrets: List[dict]):
        """ nパーティのjsonのシェア化が正しくできているかのTest """
        def dfs(val, key=""):
            dc: dict = dict()
            if isinstance(val, list):
                for v in val:
                    dc = {**dc, **dfs(v, key)}
                return dc
            if isinstance(val, dict):
                for k, v in val.items():
                    dc = {**dc, **dfs(v, key + k)}
                return dc
            return {**dc, **{key: val}}
        true_map: dict = dfs(secrets)

        for party_size in range(2, 10):
            share_json = Share.sharize(secrets, party_size=party_size)
            share_map: dict = dict()
            for share in share_json:
                for key, val in dfs(share).items():
                    if key not in share_map:
                        share_map[key] = Decimal(0)
                    share_map[key] += Decimal(val)
            for key, _ in dfs(share).items():
                assert (np.isclose(float(share_map[key]), true_map[key]))

    @pytest.mark.parametrize(
        ("secrets"),
        [
            # 1次元配列の秘密情報
            ([1e999]),
            # 2次元配列の秘密情報
            ([[1e999]])
        ]
    )
    def test_sharize_convert_to_inf(self, secrets: list):
        """ Infinityがinfに変換されているかのTest """
        for party_size in range(2, 10):
            shares = Share.sharize(secrets, party_size=party_size)

            def is_inf(val):
                ok: bool = True
                if isinstance(val, list):
                    for v in val:
                        ok &= is_inf(v)
                else:
                    ok &= (val == 'inf')
                return ok

            assert (is_inf(shares[0]))

    def test_sharize_errorhandring(self):
        """ 異常値を与えてエラーが出るかTest """
        with pytest.raises(Exception):
            Share.sharize("hey")
        with pytest.raises(Exception):
            Share.sharize(["str"])
        with pytest.raises(Exception):
            Share.sharize([[1, 2], [1]])
        with pytest.raises(Exception):
            Share.sharize([[[1]]])
        with pytest.raises(Exception):
            Share.sharize([[], []])



from typing import Iterable, List, Optional

from quickmpc import JobStatus


def make_statuses_detailed(statuses: Optional[Iterable[int]]) \
        -> Optional[List[str]]:
    if statuses is None:
        return None

    return [
        f"{JobStatus.Name(s)}({s})"
        for s in statuses
    ]



""" Testの前の前処理を定義
ローカルでhttpsサーバを3台立てており，
yieldのタイミングで全てのTestが開始
"""
import glob
import os

import pytest

from .local_server import serve


@pytest.fixture(scope='session')
def run_server1():
    """ run server """
    server = serve(1)
    server.start()

    """ test start """
    yield

    """ test end """
    for file in glob.glob("./tests/unit_tests/dim1-uuid*"):
        os.remove(file)

    server.stop(0)


@pytest.fixture(scope='session')
def run_server2():
    """ run server """
    server = serve(2)
    server.start()

    """ test start """
    yield

    """ test end """
    server.stop(0)


@pytest.fixture(scope='session')
def run_server3():
    """ run server """
    server = serve(3)
    server.start()

    """ test start """
    yield

    """ test end """
    server.stop(0)



from typing import List

import pytest

from quickmpc.exception import ArgmentError
from quickmpc.utils.make_pieces import MakePiece


class TestQMPC:

    MATRIX: List[List[str]] = [[str(i) * i for i in range(1, 6)]] * 5

    @pytest.mark.parametrize(
        ("array, size, expected"),
        [
            (["1", "2", "3", "4"], 4, [["1", "2", "3", "4"]]),
            (["1", "2", "3", "4"], 3, [["1", "2", "3"], ["4"]]),
            (["1", "2", "3", "4"], 2, [["1", "2"], ["3", "4"]]),
        ]
    )
    def test_make_pieces_array(self, array, size, expected):
        actual = MakePiece.make_pieces(array, size)
        assert (actual == expected)

    @pytest.mark.parametrize(
        ("matrix, size, expected"),
        [
            (MATRIX, 75, [MATRIX]),
            (MATRIX, 30, [MATRIX[:2], MATRIX[2:4], MATRIX[4:]]),
            (MATRIX[0], 15, [MATRIX[0]])
        ]
    )
    def test_make_pieces(self, matrix, size, expected):
        actual = MakePiece.make_pieces(matrix, size)
        assert (actual == expected)

    @pytest.mark.parametrize(
        ("string, size, expected"),
        [
            # 空文字列
            ("", 1, []),
            ("", 3, []),
            # 短めの文字列
            ("aaabbbcccddd", 6, ["aaabbb", "cccddd"]),
            ("aaabbbcccddd", 9, ["aaabbbccc", "ddd"]),
            ("aaabbbcccddd", 12, ["aaabbbcccddd"]),
            ("aaabbbcccddd", 15, ["aaabbbcccddd"]),
            # 長めの文字列
            ("a" * 1_000_000, 1, ["a" * 1] * 1_000_000),
            ("a" * 1_000_000, 1_000, ["a" * 1_000] * 1_000),
            ("a" * 1_000_000, 1_000_000, ["a" * 1_000_000] * 1),
            # 特殊な文字
            ("!\"#$%&'()=~|-^@`[]{};+:*,<.>/?", 5,
             ["!\"#$%", "&'()=", "~|-^@", "`[]{}", ";+:*,", "<.>/?"]),
        ]
    )
    def test_make_pieces_string(self, string, size, expected):
        actual = MakePiece.make_pieces(string, size)
        assert (actual == expected)

    @pytest.mark.parametrize(
        ("args, error"),
        [
            ([MATRIX, 0], RuntimeError),          # サイズが小さい場合
            ([MATRIX, 1_000_001], RuntimeError),  # サイズが大きい場合
            ([0, 0], ArgmentError),               # 引数タイプが無効な場合: 0 次元
            ([[MATRIX], 0], ArgmentError),        # 引数タイプが無効な場合: 3 次元
        ]
    )
    def test_make_pieces_errorhandring(self, args, error):
        """ 異常値を与えてエラーが出るかTest """
        with pytest.raises(error):
            MakePiece.make_pieces(*args)

    @pytest.mark.parametrize(
        ("args, error"),
        [
            (["string", 0], RuntimeError),          # サイズが小さい場合
            (["string", 1_000_001], RuntimeError),  # サイズが大きい場合
        ]
    )
    def test_make_pieces_str_errorhandring(self, args, error):
        """ 異常値を与えてエラーが出るかTest """
        with pytest.raises(error):
            MakePiece.make_pieces(*args)



import os
from typing import List

import numpy as np
import pytest

from quickmpc.utils.parse_csv import (parse, parse_csv, parse_csv_to_bitvector,
                                      parse_to_bitvector)

# 元データ
data1: List[List[str]] = [s.split(",") for s in [
    "id,attr1,attr2,attr3,attr4,attr5,attr6",
    "hoge,0,0.77,0.63,0.35,0.39,0.35",
    "huga,0,0.37,0.36,0.43,0.41,0.39",
    "piyo,1,0.34,0.34,0.44,0.50,0.32",
    "moge,1,0.47,0.43,0.34,0.29,0.34",
    "moga,0,0.67,0.41,0.25,0.49,0.25",
]]
data2: List[List[str]] = [s.split(",") for s in [
    "id,attr1,attr2,attr3",
    "hoge,0,1,3",
    "huga,0,2,1",
    "moge,1,0,4",
    "moga,0,1,2",
]]

# 正しくparseされたデータ
d1_schema: List[str] = ['id', 'attr1', 'attr2',
                        'attr3', 'attr4', 'attr5', 'attr6']
d1_secrets: List[List[float]] = [
    [230379555.4797964, 0, 0.77, 0.63, 0.35, 0.39, 0.35],
    [10723675.973257065, 0, 0.37, 0.36, 0.43, 0.41, 0.39],
    [117576607.23670769, 1, 0.34, 0.34, 0.44, 0.5, 0.32],
    [211114761.8482437, 1, 0.47, 0.43, 0.34, 0.29, 0.34],
    [13292676.303739548, 0, 0.67, 0.41, 0.25, 0.49, 0.25]
]

d2_schema: List[str] = ['id#0', 'attr1#0', 'attr1#1',
                        'attr2#0', 'attr2#1', 'attr2#2',
                        'attr3#0', 'attr3#1', 'attr3#2', 'attr3#3']
d2_secrets: List[List[float]] = [
    [230379555.4797964, 1, 0, 1, 0, 0, 1, 0, 0, 0],
    [10723675.973257065, 1, 0, 0, 1, 0, 0, 1, 0, 0],
    [211114761.8482437, 0, 1, 0, 0, 1, 0, 0, 1, 0],
    [13292676.303739548, 1, 0, 1, 0, 0, 0, 0, 0, 1]
]


def test_parse():
    """ 正しくパースできるかTest """
    secrets, schema = parse(data1)
    assert (np.allclose(secrets, d1_secrets))
    assert (schema == d1_schema)


def test_parse_to_bitvector():
    """ 正しくパースできるかTest """
    secrets, schema = parse_to_bitvector(data2, [0])
    assert (np.allclose(secrets, d2_secrets))
    assert (schema == d2_schema)


def test_parse_errorhandring():
    """ 異常値を与えてエラーが出るかTest """
    with pytest.raises(Exception):
        # 行が足りずシェアがない
        parse([["id", "a", "b", "c"]])
    with pytest.raises(Exception):
        # schemaに同じものが含まれる
        parse([["id", "a", "a"],
               ["id1", "1", "2"],
               ["id2", "3", "4"]])
    with pytest.raises(Exception):
        # 正方行列でない
        parse([["id", "a", "b"],
               ["id1", "1", "2"],
               ["id2", "3", "4", "5"]])
        parse([["id", "a", "b"],
               ["id1", "1", "2"],
               ["id2"]])


def test_parse_csv():
    """ csvを正しくパースできるかTest """
    secrets, schema = parse_csv(
        f"{os.path.dirname(__file__)}/test_files/data1.csv")
    assert (np.allclose(secrets, d1_secrets))
    assert (schema == d1_schema)


def test_parse_csv_to_bitvector():
    """ csvを正しくパースできるかTest """
    secrets, schema = parse_csv_to_bitvector(
        f"{os.path.dirname(__file__)}/test_files/data3.csv", [0])
    assert (np.allclose(secrets, d2_secrets))
    assert (schema == d2_schema)


def test_parse_csv_errorhandring():
    """ 異常値を与えてエラーが出るかTest """
    with pytest.raises(Exception):
        # fileが存在しない
        parse_csv("hoge")
    with pytest.raises(Exception):
        # formatがおかしい
        parse_csv(f"{os.path.dirname(__file__)}/test_files/data2.csv")



""" QuickMPCによる演算の動作デモ
本デモコードは次の操作を行っている．
Step 1. QuickMPC動作のための準備
Step 2. シェアをコンテナに送信
Step 3. 計算リクエストを送信
Step 4. シェアを元に復元された演算結果の真値を取得
"""

import logging
import time
from typing import Callable, List, Tuple

from utils import make_statuses_detailed

from quickmpc import QMPC

logger = logging.getLogger(__name__)
formatter = ("%(levelname)-5s | %(message)s")
logging.basicConfig(level=logging.DEBUG, format=formatter)

if __name__ == '__main__':

    """ Step 1. 諸準備 """
    # QuickMPC Setting
    qmpc: QMPC = QMPC(
        ["http://localhost:9001",
         "http://localhost:9002",
         "http://localhost:9003"]
    )
    # データ Setting
    data_dir: str = "../data/"
    secrets1, schema1 = qmpc.parse_csv_file(data_dir+"data1-1.csv")
    secrets2, schema2 = qmpc.parse_csv_file(data_dir+"data1-2.csv")
    secrets3, schema3 = qmpc.parse_csv_file(data_dir+"data-meshcode.csv")

    """ Step 2. シェア送信 """
    logger.info("---- send_share start ----")
    send_res1 = qmpc.send_share(secrets1, schema1)
    if not send_res1["is_ok"]:
        raise
    data_id1 = send_res1["data_id"]
    logger.info(f"data_id1: {data_id1}")

    time.sleep(1)
    send_res2 = qmpc.send_share(secrets2, schema2)
    if not send_res2["is_ok"]:
        raise
    data_id2 = send_res2["data_id"]
    logger.info(f"data_id2: {data_id2}")

    time.sleep(1)
    send_res3 = qmpc.send_share(secrets3, schema3)
    if not send_res3["is_ok"]:
        raise
    data_id3 = send_res3["data_id"]
    logger.info(f"data_id3: {data_id3}")
    logger.info("---- send_share success ----")

    """ Step 3. data_idに対して計算のリクエスト送信 """
    time.sleep(3)
    data_ids = [data_id1, data_id2, data_id3]
    # Jobを追加する場合はここに追加する
    # 形式: [呼び出すJob，指定する列，真値]
    exe_set: List[Tuple[Callable, List, List]] = [
        (qmpc.sum, [2, 3, 4], [12, 12, 17]),
        (qmpc.mean, [2, 3, 4], [2.4, 2.4, 3.4]),
        (qmpc.variance, [2, 3, 4], [2.24, 1.84, 4.24]),
        (
            qmpc.meshcode,
            [7, 8, 9, 10],
            [
                [53, 39, 4, 5, 4, 7, 1, 1],
                [65, 42, 5, 2, 1, 8, 3, 2],
                [47, 30, 3, 4, 0, 3, 2, 4],
                [51, 8, 4, 7, 4, 0, 3, 1],
            ]
        ),
        (qmpc.correl, [[2, 3, 4], [5]],
         [0.995537307016568, -0.192854233809008, 0.0441893193071959]),
    ]
    for exe_func, inp, true_val in exe_set:
        logger.info(f"---- {exe_func.__name__} start ----")
        # TODO: Jobごとに使用するdata_idを選択できるようにする
        exec_res = exe_func([data_ids, [0, 0], [1, 1, 1]], inp) \
            if exe_func == qmpc.meshcode  \
            else exe_func([data_ids[:2], [0], [1, 1]], inp)

        """ Step 4. 結果を取得 """
        job_uuid = exec_res["job_uuid"]
        logger.info(f"job_uuid: {job_uuid}")
        # DBに計算結果が書き込まれていないと思うので1秒待機
        time.sleep(1)

        for _ in range(100):
            # 計算が終わるか100秒(1秒x100回)経つまで繰り返す
            get_res = qmpc.get_computation_result(job_uuid)
            if get_res["results"] is not None:
                result_secrets = get_res["results"]
                logger.info(f"exec value: {result_secrets}")
                logger.info(f"true value: {true_val}")

                logger.info(f"---- {exe_func.__name__} success ----")
                break

            statuses = make_statuses_detailed(get_res['statuses'])
            logger.info(f"statuses: {statuses}")

            time.sleep(1)



from quickmpc import QMPC
from tabulate import tabulate
from typing import List
import sys


def output_table(title: int, schema: List[str], table: List):
    print("party"+str(title))
    print(tabulate(
        table,
        headers=schema,
        tablefmt="grid",
        floatfmt='.11f'))


if __name__ == '__main__':
    # QuickMPC Setting
    qmpc: QMPC = QMPC(
        ["http://localhost:9001",
         "http://localhost:9002",
         "http://localhost:9003"]
    )

    if not len(sys.argv) == 2:
        print('Usage: $ pipenv run python demo_sharize.py [file_path]',
              file=sys.stderr)
        sys.exit(1)
    # csv dataをパースする
    filename: str = sys.argv[1]

    secrets, schema = qmpc.parse_csv_file(filename)
    # データをシェア化する
    res = qmpc.demo_sharize(secrets)
    for party_id, share in enumerate(res['results']):
        output_table(party_id+1, schema, share)
